/**
 *
 * Copyright 2013 Stefan Brenner, Sven Schober
 *
 * This file is part of DAE.
 *
 * DAE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DAE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DAE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package sequoia;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.math3.stat.StatUtils;

/**
 * Represents (and calculates) a CDF for a array of values
 * 
 * @author stefan
 * 
 */
public class CDF {
	private static final int PRECISION = 1;
	private Map<Double, Double> _cdf = new HashMap<Double, Double>();
	private List<Double> _keys = new ArrayList<Double>();
	private double _mean;

	Map<Double, Double> _occurencies = new HashMap<Double, Double>();
	private int _max;

	/**
	 * Constructor takes an array of values and calculates the CDF
	 * 
	 * @param values
	 */
	public CDF(List<Double> values, int max) {
		/**
		 * In a first step, the frequency of occurence of discretized values is
		 * accumulated. Then, the cumulative distribution of these values is
		 * calculated in a second step.
		 */

		_max = max;
		
		// 1) count occurencies
		init(values);

		int totalCount = 0;
		for (double value : values) {
			// discretize values
			double val = Helper.round(value, PRECISION);
			if (Master.CDF_LIMIT)
				val = val > _max ? _max : val;

			// count frequency of occurrence
			if (!_occurencies.containsKey(val))
				_occurencies.put(val, 0D);
			_occurencies.put(val, _occurencies.get(val) + 1);
			totalCount++;
		}

		// 2a) sort values increasing
		_keys.addAll(_occurencies.keySet());
		Collections.sort(_keys);

		// 2b) calculate CDF values
		double cdfSum = 0;
		for (Double value : _keys) {
			Double frequency = _occurencies.get(value);
			double cdfValue = frequency / totalCount;
			cdfSum += cdfValue;
			_cdf.put(value, cdfSum);
		}

		_mean = StatUtils.mean(Helper.toArray(values));
	}
	
//	public CDF(double[] values, int max) {
//		List<Double> list = new ArrayList<Double>();
//		for(int i=0; i<values.length; i++)
//			list.add(values[i]);
//		this(list, max);
//	}

	/**
	 * Initialize with 0 for any occuring value in order to avoid bogus results
	 * 
	 * @param values
	 */
	private void init(List<Double> values) {
//		double[] array = Helper.toArray(values);
//		double max = Helper.round(StatUtils.max(array), PRECISION);
		double max = _max;
		double step = Math.pow(10, -PRECISION);
		for (double i = 0; i < max; i += step) {
			i = Helper.round(i,PRECISION);
			_occurencies.put(i, 0D);
		}
	}

	public Map<Double, Double> get_cdf() {
		return _cdf;
	}

	public List<Double> get_keys() {
		return _keys;
	}

	public double get(double key) {
		return _cdf.get(key);
	}

	public double getMean() {
		return _mean;
	}

	public String toString() {
		String out = "";
		String FORMAT = "%s";
		for (Double key : get_keys()) {
			out += String.format(FORMAT + ";" + FORMAT + "\n", key, get(key));
		}
		return out;
	}
}
