/**
 *
 * Copyright 2013 Stefan Brenner, Sven Schober
 *
 * This file is part of DAE.
 *
 * DAE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DAE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DAE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package sequoia.topology;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import sequoia.Master;

/**
 * The PlanetLab topology, based on a download of the dataset at the beginning
 * of the working period on the thesis.
 */
public class TopPlanetLab extends Topology {

	// activate Clique-only mode (else: full dataset)
	private static boolean USE_CLIQUE = false;

	// list of hosts, that are part of the clique
	private List<String> _clique_hosts = new ArrayList<String>();

	private boolean skippedvalues = false;
	
	/**
	 * init directly via original dataset file
	 * format: hosta,hostb,-,-,capacity,pathchirp,spruce, ...
	 *         0     1     2 3 4        5         6
	 */
	public TopPlanetLab() {

		try {
			// read complete file
			BufferedReader br = new BufferedReader(new FileReader("data/hp-planetlab-bandwidth-snapshot.csv"));
			String linebuffer;
			while ((linebuffer = br.readLine()) != null) {

				// skip comments
				linebuffer = linebuffer.trim();
				if (linebuffer.startsWith("#"))
					continue;

				// parse line
				String[] parts = linebuffer.split(",");
				if (parts.length < 6)
					continue;

				double bw_spruce = 0;
				if (!parts[4].equals("N/A"))
					bw_spruce = Double.parseDouble(parts[4]);

				// filter
				if(bw_spruce > Master.CONST && !skippedvalues) {
					System.err.println("Very high bandwidth value skipped... "+bw_spruce);
					skippedvalues = true;
					continue;
				}
				
				// store
				String hostname_a = parts[0];
				String hostname_b = parts[1];
				int availBw = (int) Math.round(bw_spruce);

				if(availBw <= 0)
					continue;
				
				// store new hostnames
				rememberHostname(hostname_a);
				rememberHostname(hostname_b);

				// store value
				addLinkBandwidth(hostname_a, hostname_b, availBw);
			}

			br.close();
			shuffle();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * init via filtered file
	 * format: hosta,hostb,measurement
	 */
	public TopPlanetLab(String filename) {
		
		try {
			// read complete file
			BufferedReader br = new BufferedReader(new FileReader("data/"+filename));
			String linebuffer;
			while ((linebuffer = br.readLine()) != null) {
				String[] parts = linebuffer.split(",");
				if (parts.length < 3)
					continue;
				String hostname_a = parts[0];
				String hostname_b = parts[1];
				double bw = Double.parseDouble(parts[2]);
				rememberHostname(hostname_a);
				rememberHostname(hostname_b);
				addLinkBandwidth(hostname_a,hostname_b,(int)Math.round(bw));
			}
			br.close();
			shuffle();
		} catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Read clique hosts from file
	 */
	public void readCliqueFile(String filename) {
		try {
			BufferedReader br = new BufferedReader(new FileReader(filename));

			String linebuffer;
			while ((linebuffer = br.readLine()) != null) {

				// skip comments
				linebuffer = linebuffer.trim();
				if (linebuffer.startsWith("#"))
					continue;

				// whole line is a hostname
				if (!_clique_hosts.contains(linebuffer)) {
					_clique_hosts.add(linebuffer);
				}
			}

			br.close();

			USE_CLIQUE = true;
			Collections.shuffle(_clique_hosts, rnd);

		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public List<String> getAllHosts() {
		List<String> list;
		if (USE_CLIQUE) {
			list = getAllHostsClique();
		} else {
			list = super.getAllHosts();
		}
		return list;
	}

	public List<String> getAllHostsClique() {
		List<String> list = new ArrayList<String>(_clique_hosts);
		Collections.copy(list, _clique_hosts);
		return list;
	}

	public String toString() {
		return super.toString();
	}

	@SuppressWarnings("unused")
	private String hostnameFormat(String s) {
		return s.substring(0, 14);
	}

	public void shuffle() {
		if (USE_CLIQUE) {
			Collections.shuffle(_clique_hosts, rnd);
		} else {
			super.shuffle();
		}
	}
}
