/**
 *
 * Copyright 2013 Stefan Brenner, Sven Schober
 *
 * This file is part of DAE.
 *
 * DAE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DAE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DAE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package sequoia.tree;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import sequoia.Master;
import sequoia.Pair;
import sequoia.topology.ITopology;
import sequoia.tree.AsymHost.Direction;

/**
 * Duplicate embedding prediction tree
 */
public class TreeMy extends PredictionTree<AsymHost> {

	protected List<AsymHost> _hosts = new ArrayList<AsymHost>();
	protected String _firstHostname;

	private boolean smallvalue = false;
	private boolean highvalue = false;
	private int counter = 0;
	
	private Map<String, List<String>> _measured_out = new HashMap<String, List<String>>();
	private Map<String, List<String>> _measured_in = new HashMap<String, List<String>>();
	
	private static final double NA_DIST = Master.CONST;
	private static final double NA_BW = 0;
	
	public TreeMy(ITopology topology) {
		_topology = topology;
	}

	/** choose embedding algorithm according to current node count of the tree */
	public void embed(String joiner) throws RuntimeException {
		if (_hosts.contains(joiner))
			return;

		counter++;
		
		if (_hosts.size() <= 0) {
			firstJoin(joiner);
		} else if (_hosts.size() <= 2) {
			secondJoin(joiner);
		} else {
			anyJoin(joiner);
		}

		// verbose: status output
		int count = ((_hosts.size() - 1) / 2 + 1);
		if (Master.cli_verbose && count % 10 == 0)
			System.err.printf("Tried: %s hosts; Embedded: %s hosts\n",counter,count);
	}

	/**
	 * First node joining procedure. 
	 */
	protected void firstJoin(String hostname) {
		_firstHostname = hostname;
		AsymHost host = new AsymHost(hostname);
		host.label.add(host, 0, 0);
		_hosts.add(host);
	}

	/**
	 * Second node joining procedure
	 */
	protected void secondJoin(String joinerHostname) {
		AsymHost base = _hosts.get(0);
		double up = measureDistance(joinerHostname, _hosts.get(0).hostname, Direction.UPSTREAM);
		double down = measureDistance(joinerHostname, _hosts.get(0).hostname, Direction.DOWNSTREAM);
		
		if(Double.isInfinite(up) || Double.isInfinite(down))
			throw new RuntimeException("Second host one or both measurements missing");
		
		int var = 1;
		
		// #1 (thesis version)
		// firstly embed the instance with greater distance (i.e. smaller bandwidth)
		if(var==1) {
			if (up > down) {
				AsymHost upstream = createHost(joinerHostname, Direction.UPSTREAM, base, 0, up);
				_hosts.add(upstream);
				AsymHost downstream = createHost(joinerHostname, Direction.DOWNSTREAM, upstream, up - down / 2, down / 2);
				_hosts.add(downstream);
			} else {
				AsymHost downstream = createHost(joinerHostname, Direction.DOWNSTREAM, base, 0, down);
				_hosts.add(downstream);
				AsymHost upstream = createHost(joinerHostname, Direction.UPSTREAM, downstream, down - up / 2, up / 2);
				_hosts.add(upstream);
			}
		} else if(var==2) {
			// #2 
			AsymHost upstream = createHost(joinerHostname, Direction.UPSTREAM, base, 0, up);
			_hosts.add(upstream);
			AsymHost downstream = createHost(joinerHostname, Direction.DOWNSTREAM, upstream, up / 2, down - up / 2);
			_hosts.add(downstream);
		} else if(var==3) {
			// #3
			AsymHost downstream = createHost(joinerHostname, Direction.DOWNSTREAM, base, 0, down);
			_hosts.add(downstream);
			AsymHost upstream = createHost(joinerHostname, Direction.UPSTREAM, downstream, down / 2, up - down / 2);
			_hosts.add(upstream);
		}
	}

	/**
	 * Regular node joining procedure
	 */
	protected void anyJoin(String joiner) {
		
		AsymHost base_u;
		AsymHost base_d;
		AsymHost anchor_u;
		AsymHost anchor_d;
		
		int tries = 0;
		while(true) {
		
			// try finding a basenode/anchor with direction opposite to joiner
			base_d = getDirectedRandomNode(Direction.DOWNSTREAM);
			anchor_d = getAnchor(base_d, joiner, Direction.DOWNSTREAM);
			base_u = getDirectedRandomNode(Direction.UPSTREAM);
			anchor_u = getAnchor(base_u, joiner, Direction.UPSTREAM);
	
			// optimize base/anchor by error minimization
			if (Master.cli_optimize) {
				Pair<AsymHost> pair_d = optimize(joiner, Direction.UPSTREAM, anchor_d, base_d);
				Pair<AsymHost> pair_u = optimize(joiner, Direction.DOWNSTREAM, anchor_u, base_u);
				if(pair_d != null && (!anchor_d.equals(pair_d.getLeft()) || !base_d.equals(pair_d.getRight()))) {
					anchor_d = pair_d.getLeft();
					base_d = pair_d.getRight();
				}
				if(pair_u != null && (!anchor_u.equals(pair_u.getLeft()) || !base_u.equals(pair_u.getRight()))) {
					anchor_u = pair_u.getLeft();
					base_u = pair_u.getRight();
				}
	 		}
		
			if (anchor_u != null && anchor_d != null) {
				break;
			} else {
				// try some multiple times then quit
				if(tries++ > 10) { throw new RuntimeException("No reference nodes found for given joiner: "+joiner); }
			}
		}

		// actual embedding
		AsymHost joinerHost_d = position(joiner, Direction.DOWNSTREAM, anchor_u, base_u);
		AsymHost joinerHost_u = position(joiner, Direction.UPSTREAM, anchor_d, base_d);
		_hosts.add(joinerHost_d);
		_hosts.add(joinerHost_u);
	}

	protected AsymHost getRandomNode() {
		int idx = (int) Math.round(Master.rand.nextDouble() * (_hosts.size() - 1));
		return _hosts.get(idx);
	}

	protected AsymHost getDirectedRandomNode(Direction dir) {
		AsymHost node = null;
		while (true) {
			node = getRandomNode();
			if (node.direction == null || node.direction.equals(dir)) {
				return node;
			}
		}
	}

	// Find anchor of direction anchorDir for joiner using basenode. 
	protected AsymHost getAnchor(AsymHost basenode, String joiner, Direction anchorDir) {
		
		Direction joinerDir = anchorDir.reverse();
		AsymHost anchor = null;
		while (true) {
			double min_error = Double.POSITIVE_INFINITY;
			AsymHost min_host = null;
			
			List<AsymHost> candidates = getCandidates(basenode, 2);
			for (AsymHost anchor_cand : candidates) {
				if (!anchorIsValid(joiner, joinerDir, anchor_cand, basenode))
					continue;
				
				try {
					AsymHost temp = position(joiner, joinerDir, anchor_cand, basenode);
					double error = relError(temp,basenode);
					if (error < min_error) {
						min_error = error;
						min_host = anchor_cand;
					}
					if (min_error <= 0 && min_host != null)
						break;
					
				} catch(Exception e) {
					if(Master.cli_debug)
						System.err.println(e.getMessage());
				}
			}
			
			if (min_host == null)
				break;

			if (anchor != null && min_host.equals(anchor))
				break;

			anchor = min_host;
		}
		return anchor;
	}

	/**
	 * Get candidate, with iteration parameter, in order to recursively create
	 * candidates of two-hop neighborhood
	 */
	protected List<AsymHost> getCandidates(AsymHost basenode, int i) {
		List<AsymHost> candidates = new ArrayList<AsymHost>();
		if (i > 0) {
			if (basenode.parent != null) {
				for (AsymHost h : getCandidates(basenode.parent, i - 1)) {
					if (!candidates.contains(h))
						candidates.add(h);
				}
			}
			for (AsymHost child : basenode.children) {
				for (AsymHost h : getCandidates(child, i - 1)) {
					if (!candidates.contains(h))
						candidates.add(h);
				}
			}
		}

		if (!candidates.contains(basenode))
			candidates.add(basenode);
		return candidates;
	}

	/**
	 * check anchor validity
	 */
	protected boolean anchorIsValid(String joinerHost, Direction joinerDir, AsymHost anchor, AsymHost base) {

		// basic check
		if (anchor == null || base == null) {
			return false;
		}

		// anchor shall not be first embedded node
		if (anchor.hostname.equals(_firstHostname)) {
			return false;
		}

		// anchor must have opposite direction to joiner
		if (anchor.direction.equals(joinerDir)) {
			return false;
		}

		// hostnames must be different: joiner, base, anchor
		String anchorHost = anchor.hostname;
		String baseHost = base.hostname;
		if (joinerHost.equals(anchorHost) || joinerHost.equals(baseHost) || anchorHost.equals(baseHost)) {
			return false;
		}

		return true;
	}

	/**
	 * Check if base-anchor-combination is valid
	 */
	protected boolean baseAnchorIsValid(String joiner, Direction joinerDir, AsymHost anchor, AsymHost base) {
		// basic check
		if (anchor == null || base == null) {
			return false;
		}

		// anchor shall not be first embedded node
		if (anchor.hostname.equals(_firstHostname)) {
			return false;
		}

		// anchor must have opposite direction to joiner
		if (anchor.direction.equals(joinerDir)) {
			return false;
		}

		if (base.direction != null && base.direction.equals(joinerDir)) {
			return false;
		}

		return true;
	}

	/**
	 * Optimization of base and anchor. Finally not used.
	 */
	protected Pair<AsymHost> optimize(String joiner, Direction joinerDir, AsymHost initialAnchor, AsymHost initialBase) {
		if(initialAnchor == null || initialBase == null)
			return null;
		
		// determine whether or not optimization might improve accuracy
		AsymHost tmp = position(joiner, joinerDir, initialAnchor, initialBase);
		double initialError = relError(tmp,initialBase);
		if (initialError <= 0 && Master.cli_debug) {
			System.err.println("Yey I think that might be shit what I should do here... ");
		}

		// do actual optimization
		List<String> measured = getMeasuredList(joiner, joinerDir);
		AsymHost optBase = null;
		AsymHost optAnchor = null;
		double min_error = initialError;

		if (measured == null)
			return null;

		outer: for (String b : measured) {
			AsymHost base = getHost(b, joinerDir.reverse());
			
			for (String a : measured) {
				if (a.equals(b))
					continue;

				AsymHost anchor = getHost(a, joinerDir.reverse());

				if (!baseAnchorIsValid(joiner, joinerDir, base, anchor))
					continue;

				AsymHost temp = position(joiner, joinerDir, anchor, base);
				double error = relError(temp,base);
				if (error < min_error) {
					min_error = error;
					optBase = base;
					optAnchor = anchor;
				}

				if (min_error <= 0)
					break outer;
			}
		}

		if (optBase != null && optAnchor != null) {
			Pair<AsymHost> pair = new Pair<AsymHost>(optAnchor, optBase);
			return pair;
		}

		return null;
	}

	/**
	 * Position a host on the tree.
	 */
	protected AsymHost position(String joiner, Direction joinerDir, AsymHost anchor, AsymHost base) {
		
		// measure
		double distance_joiner_base = measureDistance(joiner, base.hostname, joinerDir);
		double distance_joiner_anchor = measureDistance(joiner, anchor.hostname, joinerDir);
		
		if(Double.isInfinite(distance_joiner_base) || Double.isInfinite(distance_joiner_anchor))
			throw new RuntimeException("Unusable distance for positioning");
		
		if(Math.abs(distance_joiner_base - Master.CONST)<0.01 || Math.abs(distance_joiner_anchor - Master.CONST) < 0.01) {
			throw new RuntimeException("Unusable distance for positioning");
		}
		
		double distance_anchor_base = distAnchorBase(anchor, base);
		
		// calculate xz yz xy
		double gromov = gromovProduct(distance_joiner_base, distance_anchor_base, distance_joiner_anchor);
		double offset = distance_anchor_base - gromov;
		double distance = distance_joiner_base - gromov;
		
		int s = anchor.label.distances.size(); 
	 	if( s > 1){
		 	Double anchor_distance = anchor.label.distances.get(s - 2); 
		 	if( offset > anchor_distance){
		 		// System.err.println("limiting offset");
			 	offset = anchor_distance;
		 	}
		 	 
	 	}
		
		if(Master.cli_debug && (Double.isInfinite(offset) || Double.isInfinite(distance)))
			System.err.println("Infinity offset or distance"+offset+" "+distance);
		
		if(Master.cli_debug && (Math.abs(distance) < 0.001 || Math.abs(offset) < 0.001)) 
			System.err.println("Small distance or offset");
		
		AsymHost host = createHost(joiner, joinerDir, anchor, offset, distance);
		return host;
		
	}

	/**
	 * Distance between anchor and base
	 */
	protected double distAnchorBase(AsymHost anchor, AsymHost base) {
		double a = anchor.label.distanceTo(base.label);
		return a;
	}

	/**
	 * Create AsymHost instance which is added to the tree finally
	 */
	protected AsymHost createHost(String hostname, Direction hostDir, AsymHost anchor, double offset, double distance) {
		AsymHost host = new AsymHost(hostname, hostDir);
		host.label = anchor.label.clone();
		host.label.updateLast(offset, distance);
		host.label.add(host, 0, 0);
		host.parent = anchor;
		host.parent.children.add(host);
		return host;
	}

	/**
	 * Calculate relative prediction error for a temporary position
	 */
	protected double relError(AsymHost tempPosition,AsymHost baseNode) {
		
		List<String> measuredHosts = getMeasuredList(tempPosition.hostname, tempPosition.direction);
		double sum = 0;
		int count = 0;
		for (String hostname : measuredHosts) {
			AsymHost host = new AsymHost(hostname, tempPosition.direction.reverse());
			DistanceLabel<AsymHost> label = getDLabel(host.toString());
			if(label == null)
				continue;
			double measuredDistance = measureDistance(tempPosition.hostname, hostname, tempPosition.direction);
			if(Double.isInfinite(measuredDistance))
				continue;
			double measuredBandwidth = transform(measuredDistance);
			double predictedDistance = tempPosition.label.distanceTo(label);
			double predictedBandwidth = transform(predictedDistance);
			double diff = Math.abs(predictedBandwidth - measuredBandwidth);
			double error = diff / measuredBandwidth;
			sum += error;
			count++;
		}
		double error = sum / count;
		if(Double.isInfinite(error) || Double.isNaN(error))
			error = Double.POSITIVE_INFINITY;
		
		return error;
	}

	protected AsymHost getHost(String hostname, Direction dir) {
		int idx = _hosts.indexOf(new AsymHost(hostname, dir));
		if(idx == -1)
			return null;
		return _hosts.get(idx);
	}

	@Override
	public double transform(double v) {
		if(Master.cli_debug && v > Master.CONST && !highvalue) {
			highvalue = true;
			System.err.println("Transform value greater than CONST "+v);
		}
		if(Master.cli_debug && Math.abs(v) < 0.001 && !smallvalue) {
			smallvalue = true;
			System.err.println("Transform value very small "+v);
		}
		if(Math.abs(v) < 0.001)
			v = 0.001;
		return Master.CONST / v;
	}

	@Override
	public List<String> getHostnames() {
		ArrayList<AsymHost> hosts = new ArrayList<AsymHost>(_hosts);
		Collections.copy(hosts, _hosts);
		Collections.sort(hosts, new AsymHostLabelComparator());
		List<String> list = new ArrayList<String>();
		for (AsymHost host : hosts) {
			if (!list.contains(host.hostname))
				list.add(host.hostname);
		}
		return list;
	}

	public List<AsymHost> getHosts() {
		List<AsymHost> list = new ArrayList<AsymHost>(_hosts);
		Collections.copy(list, _hosts);
		Collections.sort(list, new AsymHostLabelComparator());
		return list;
	}

	@Override
	public DistanceLabel<AsymHost> getDLabel(String host) {
		String[] tokens = host.split(":");
		if (tokens.length == 2) {
			Direction dir = tokens[1].equals("U") ? Direction.UPSTREAM : Direction.DOWNSTREAM;
			AsymHost search = new AsymHost(tokens[0], dir);
			if (_hosts.contains(search))
				return _hosts.get(_hosts.indexOf(search)).label;
		} else {
			if (_hosts.get(0).hostname.startsWith(host))
				return _hosts.get(0).label;
		}
		return null;
	}

	@Override
	public double predictBandwidth(String a, String b) {
		if (a.equals(b))
			return 0;

		AsymHost ah = new AsymHost(a, Direction.UPSTREAM);
		AsymHost bh = new AsymHost(b, Direction.DOWNSTREAM);
		DistanceLabel<AsymHost> upstream = getDLabel(ah.toString());
		DistanceLabel<AsymHost> downstream = getDLabel(bh.toString());
		if(upstream == null || downstream == null)
			return NA_DIST; 
		double distance = upstream.distanceTo(downstream);
		double bandwidth = transform(distance);
		if(Math.abs(Master.CONST - bandwidth) < 1 || Math.abs(Master.CONST - distance) < 1) {
			System.err.println("Infinity");
		}
		return bandwidth;
	}

	protected double measureDistance(String a, String b, Direction dir) {
		if (dir.equals(Direction.UPSTREAM))
			return measureDistance(a, b);
		else
			return measureDistance(b, a);
	}

	protected double measureDistance(String a, String b) {
		double bw = measureBandwidth(a,b);
		double dist = transform(bw);
		return dist;
	}

	protected double measureBandwidth(String a, String b) {
		double bw = _topology.getBandwidth(a, b);
		if(bw != NA_BW) 
			countMeasurement(a, b);
		return bw;
	}
	
	private void countMeasurement(String a, String b) {
		// remember outgoing measurement from a to b
		if(!_measured_out.containsKey(a))
			_measured_out.put(a,new ArrayList<String>());
		if(!_measured_out.get(a).contains(b))
			_measured_out.get(a).add(b);
		// remember incoming measurement to b coming from a
		if(!_measured_in.containsKey(b))
			_measured_in.put(b,new ArrayList<String>());
		if(!_measured_in.get(b).contains(a))
			_measured_in.get(b).add(a);
 	}
	
	public double getMeasurementRatio(){
		double measurements = 0;
		List<String> hostnames = getHostnames();
		// for all tree hosts
		for(AsymHost host : _hosts) {
			// count measurements originating from them to other hosts available on the tree
			if(_measured_out.containsKey(host.hostname)) {
				List<String> measuredList = _measured_out.get(host.hostname);
				for(String destinationHostname : measuredList) {
					if(hostnames.contains(destinationHostname))
						measurements++;
				}
			}
		}
		double size = hostnames.size();
		double links = (size * (size - 1));
		double ratio = measurements / links;
		return ratio;
	}
	
	protected List<String> getMeasuredList(String hostname, Direction dir) {
		if (dir.equals(Direction.UPSTREAM)) {
			if(!_measured_out.containsKey(hostname))
				_measured_out.put(hostname, new ArrayList<String>());
			return _measured_out.get(hostname);
		} else {
			if(!_measured_in.containsKey(hostname))
				_measured_in.put(hostname, new ArrayList<String>());
			return _measured_in.get(hostname);
		}
	}

	public String toString() {
		List<AsymHost> hosts = new ArrayList<AsymHost>(_hosts);
		Collections.copy(hosts, _hosts);
		Collections.sort(hosts, new AsymHostLabelComparator());
		String s = "Tree Structure of "+ hosts.size() + " nodes\n";
		for (AsymHost h : hosts)
			s += h + String.format("%s\n", h.label);
		return s;
	}
	
}
