/**
 *
 * Copyright 2013 Stefan Brenner, Sven Schober
 *
 * This file is part of DAE.
 *
 * DAE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DAE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DAE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package sequoia.tree;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import sequoia.Master;
import sequoia.topology.ITopology;

/**
 * Simulates the Sequoia approach of prediction trees.
 */
public class TreeSequoia extends PredictionTree<Host> {

	private static final double NA_BW = 0;
	private static final double NA_DIST = Master.CONST;
	protected List<Host> _hosts = new ArrayList<Host>();
	protected Host _lever;
	protected Map<String, List<String>> _measured = new HashMap<String, List<String>>();
	private boolean valgtconst = false;
	
	public TreeSequoia(ITopology top) {
		_topology = top;
	}

	public void embed(String newHost) {
		if (_hosts.contains(newHost))
			return;
		else if (_hosts.size() == 0)
			firstJoin(newHost);
		else if (_hosts.size() == 1)
			secondJoin(newHost);
		else
			anyJoin(newHost);
	}

	public List<String> getHostnames() {
		ArrayList<Host> hosts = new ArrayList<Host>(_hosts);
		Collections.copy(hosts, _hosts);
		Collections.sort(hosts, new HostLabelComparator());

		List<String> list = new ArrayList<String>();
		for (Host h : hosts)
			list.add(h.hostname);

		return list;
	}

	@Override
	@SuppressWarnings({ "rawtypes", "unchecked" })
	public DistanceLabel getDLabel(String host) {
		return getLabel(host).clone();
	}

	public double predictBandwidth(String a, String b) {
		double distance = predictDistance(a, b);
		double bandwidth = transform(distance);
		if(bandwidth < 0 && Master.cli_debug)
			System.err.println("negative bandwidth prediction");
		return bandwidth;
	}

	public double transform(double v) {
		if(v > Master.CONST && !valgtconst) {
			valgtconst = true;
			if(Master.cli_debug)
				System.err.println("transform: Value is greater than CONST: "+v+" oversize: "+(Master.CONST-v) + " [...]");
		}
		double val = Master.CONST - v;
		return val;
	}

	@Override
	public List<Host> getHosts() {
		ArrayList<Host> hosts = new ArrayList<Host>(_hosts);
		Collections.copy(hosts, _hosts);
		return hosts;
	}

	/**
	 * First host joining procedure
	 */
	protected void firstJoin(String hostname) {
		Host joiner = new Host(hostname);
		joiner.parent = null;
		joiner.label.add(joiner, 0, 0);
		_lever = joiner;
		_hosts.add(joiner);
	}

	/**
	 * Second host joining procedure
	 */
	protected void secondJoin(String hostname) {
		Host joiner = new Host(hostname);

		// if measurement fails, skip node embedding
		double distance = measureDistance(joiner, _hosts.get(0));
		if (distance == NA_DIST) {
			if(Master.cli_debug)
				System.err.print("skipped second host ");
			return;
		}
		
		joiner.parent = _hosts.get(0);
		joiner.parent.children.add(joiner);
		joiner.label.add(joiner.parent, 0, distance);
		joiner.label.add(joiner, 0, 0);
		_hosts.add(joiner);
	}

	/**
	 * Join procedure for regular nodes
	 */
	protected void anyJoin(String hostname) {
		// any node join algorithm
		Host joiner = new Host(hostname);
		Host lever = _hosts.get(0);
		Host anchor = getAnchor(lever, joiner);

		// check conditions to continue
		if (anchor != null && lever != null && !anchor.equals(lever) && !anchor.equals(_lever)) {
			try {
				embedHost(joiner, anchor, lever);
			} catch(Exception e) {
				if(Master.cli_debug) {
					System.err.println("anyJoin(): skipped a regular host");
					e.printStackTrace();
				}
			}
		} 
		
		if (Master.cli_verbose && _hosts.size() % 10 == 0)
			System.err.print(_hosts.size()+" ");
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	private double predictDistance(String a, String b) {
		try {
			DistanceLabel labela = getLabel(a);
			DistanceLabel labelb = getLabel(b);
			double distance = labela.distanceTo(labelb);
			return distance;
		} catch(Exception e) { 
			return NA_DIST; 
		}
	}

	private void embedHost(Host joiner, Host anchor, Host lever) {
		// get distances
		double distance_joiner_lever = measureDistance(joiner, lever);
		double distance_anchor_lever = predictDistance(anchor.hostname, lever.hostname);
		double distance_joiner_anchor = measureDistance(joiner, anchor);

		// gromov = -Inf, when measurement fails.
		if (distance_joiner_anchor == NA_DIST) {
			throw new RuntimeException("Measurement between joiner and anchor missing.");
		}

		if (distance_joiner_lever == NA_DIST) {
			throw new RuntimeException("Measurement between joiner and lever missing.");
		}

		// calculate distance label
		double gromov_product = gromovProduct(distance_joiner_lever, distance_anchor_lever, distance_joiner_anchor);
		double offset = distance_anchor_lever - gromov_product;
		double distance = distance_joiner_lever - gromov_product;

		// create distance label
		joiner.label = anchor.label.clone();
		joiner.label.updateLast(offset, distance);
		joiner.label.add(joiner, 0, 0);

		// parent relationship
		anchor.children.add(joiner);
		joiner.parent = anchor;

		// add node to tree
		_hosts.add(joiner);
	}

	protected Host getAnchor(Host lever, Host joiner) {

		double max = Double.NEGATIVE_INFINITY;
		Host anchor = null;
		Host base = lever;

		while (true) {
			Host local = null;
			List<Host> cands = getCandidates(base);
			for (Host anchor_cand : cands) {
				try {
					if (anchor_cand == null || anchor_cand.equals(lever))
						continue;
	
					double dist_anchor_lever = predictDistance(anchor_cand.hostname, lever.hostname);
					double dist_joiner_anchor = measureDistance(joiner, anchor_cand);
					double dist_joiner_lever = measureDistance(joiner, lever);
					
					double gromov = 0;
					if(dist_joiner_anchor == NA_DIST || dist_joiner_lever == NA_DIST)
						continue; // equals to : gromov = Double.NEGATIVE_INFINITY;
					else 
						gromov = gromovProduct(dist_joiner_lever, dist_anchor_lever, dist_joiner_anchor);
					
					if (gromov > max) {
						max = gromov;
						local = anchor_cand;
					}
				} catch(Exception e) {
					System.err.println(e.getMessage());
				}
			}

			if (local == null) {
				break;
			}

			// repeat until no progress can be made
			if (local.equals(base)) {
				break;
			} else {
				base = local;
				anchor = local;
			}
		}

		return anchor;
	}

	protected List<Host> getCandidates(Host base) {
		// candidates are all leaf nodes, without lever
		List<Host> list = new ArrayList<Host>();

		list.add(base);

		for (Host child : base.children) {
			if (child != null)
				list.add(child);
		}

		if (base.parent != null)
			list.add(base.parent);

		list.remove(_hosts.get(0)); // remove lever
		return list;
	}

	protected Host getHost(Host host) {
		int idx = _hosts.indexOf(host);
		if (idx >= 0) {
			return _hosts.get(idx);
		} else {
			throw new RuntimeException("Host not found: getHost(" + host + ")");
		}
	}

	@SuppressWarnings("rawtypes")
	protected DistanceLabel getLabel(String host) {
		return getHost(new Host(host)).label;
	}

	protected double measureDistance(Host a, Host b) {
		return measureDistance(a.hostname, b.hostname);
	}

	protected double measureDistance(String a, String b) {
		double bandwidth = measureBandwidth(a,b);
		double distance = transform(bandwidth);
		return distance;
	}
	
	protected double measureBandwidth(String a, String b) {
		double bandwidth_forth = (_topology.getBandwidth(a, b));
		double bandwidth_back = (_topology.getBandwidth(b, a));
		
		if(bandwidth_forth == 0 && bandwidth_back == 0)
			return NA_BW;
		
		countMeasurement(a,b);
		
		if(bandwidth_forth == 0 && bandwidth_back != 0)
			bandwidth_forth = bandwidth_back;
		if(bandwidth_back == 0 && bandwidth_forth != 0)
			bandwidth_back = bandwidth_forth;

		double bandwidth = (bandwidth_forth + bandwidth_back) / 2;
		return bandwidth;
	}
	
	protected void countMeasurement(String a, String b) {
		if(!_measured.containsKey(a))
			_measured.put(a, new ArrayList<String>());
		if(!_measured.get(a).contains(b))
			_measured.get(a).add(b);
		if(!_measured.containsKey(b))
			_measured.put(b, new ArrayList<String>());
		if(!_measured.get(b).contains(a))
			_measured.get(b).add(a);
	}

	public double getMeasurementRatio(){
		double measurements = 0;
		List<String> hostnames = getHostnames();
		// for all tree hosts
		for(Host host : _hosts) {
			// count measurements originating from them to other hosts available on the tree
			if(_measured.containsKey(host.hostname)) {
				List<String> measuredList = _measured.get(host.hostname);
				for(String destinationHostname : measuredList) {
					if(hostnames.contains(destinationHostname))
						measurements++;
				}
			}
		}
		double size = _hosts.size();
		double links = (size * (size - 1));
		double ratio = measurements / links;
		return ratio;
	}
	
	private String hostnameFormat(String s) {
		if (s.length() > 14)
			return s.substring(0, 14);
		else
			return s;
	}

	public String toString() {
		ArrayList<Host> hosts2 = new ArrayList<Host>(_hosts);
		Collections.copy(hosts2, _hosts);
		Collections.sort(hosts2, new HostLabelComparator());
	
		String result = "Tree Structure of "+ hosts2.size() + " nodes\n";
		for (Host h : hosts2) {
			result += String.format("%10s\t%s\n", hostnameFormat(h.toString()), h.label);
		}
	
		return result;
	}

}
